@@ -1,6 +1,5 @@
 import collections
 import contextlib
-import filecmp
 import os
 import re
 import shutil
@@ -68,15 +67,45 @@ def _remove_ply_patch_annotation(lines):
         del lines[idx]
 
 
+def _remove_trailing_extra_blank_lines_from_subject(lines):
+    """Different versions of git will output different amounts of trailing
+    whitespace at the end of a subject headers, so we normalize it by only
+    allowing a single trailing blank line.
+    """
+    for idx, line in enumerate(lines):
+        if line.startswith('diff --git'):
+            break
+    else:
+        return
+
+    if idx < 2:
+        return
+
+    # If we have two blanks above first `diff --git` remove one of them
+    if not lines[idx - 1].strip() and not lines[idx - 2].strip():
+        del lines[idx - 1]
+
+
 def _fixup_patch(from_file, to_file):
     lines = from_file.readlines()
     _replace_from_sha1(lines)
     _replace_git_version(lines)
     _remove_ply_patch_annotation(lines)
+    _remove_trailing_extra_blank_lines_from_subject(lines)
     to_file.write(''.join(lines))
 
 
-class WorkingRepo(git.Repo):
+
+class Repo(git.Repo):
+    def _add_annotation(self, prefix, value):
+        """Add annotation to the last commit."""
+        commit_msg = self.log(count=1, pretty='%B')
+        if prefix not in commit_msg:
+            annotation = '%s: %s' % (prefix, value)
+            self.commit(msgs=[commit_msg, annotation], amend=True)
+
+
+class WorkingRepo(Repo):
     """Represents our local fork of the upstream repository.
 
     This is where we will create new patches (save) or apply previous patches
@@ -86,10 +115,7 @@ class WorkingRepo(git.Repo):
 
     def _add_patch_annotation(self, patch_name):
         """Add a patch annotation to the last commit."""
-        commit_msg = self.log(count=1, pretty='%B')
-        if 'Ply-Patch' not in commit_msg:
-            commit_msg += '\n\nPly-Patch: %s' % patch_name
-            self.commit(commit_msg, amend=True)
+        self._add_annotation('Ply-Patch', patch_name)
 
     def _get_patch_annotation(self, description):
         """Return the Ply-Patch annotation if present.
@@ -175,21 +201,6 @@ class WorkingRepo(git.Repo):
 
         return applied
 
-    def _store_patch_files(self, patch_names, filenames,
-                           parent_patch_name=None):
-        """Store a set of patch files in the patch-repo."""
-        source_paths = [os.path.join(self.path, f) for f in filenames]
-        return self.patch_repo.add_patches(
-            patch_names, source_paths, parent_patch_name=parent_patch_name)
-
-    def _commit_to_patch_repo(self, commit_msg):
-        if not self.patch_repo.uncommitted_changes():
-            return
-
-        based_on = self._last_upstream_commit_hash()
-        commit_msg += '\n\nPly-Based-On: %s' % based_on
-        self.patch_repo.commit(commit_msg)
-
     @property
     def patch_repo_path(self):
         try:
@@ -203,12 +214,15 @@ class WorkingRepo(git.Repo):
         """Return a patch repo object associated with this working repo via
         the ply.patchrepo git config.
         """
-        if not self.patch_repo_path:
-            raise exc.NoLinkedPatchRepo
+        if not hasattr(self, '_patch_repo'):
+            if not self.patch_repo_path:
+                raise exc.NoLinkedPatchRepo
 
-        return PatchRepo(self.patch_repo_path,
-                         quiet=self.quiet,
-                         supress_warnings=self.supress_warnings)
+            self._patch_repo =  PatchRepo(
+                             self.patch_repo_path,
+                             quiet=self.quiet,
+                             supress_warnings=self.supress_warnings)
+        return self._patch_repo
 
     @property
     def _patch_conflict_path(self):
@@ -285,7 +299,7 @@ class WorkingRepo(git.Repo):
         change was made upstream.
         """
         patch_name = self._resolve_conflict('skip')
-        self.patch_repo.remove_patches([patch_name])
+        self.patch_repo.remove_patch(patch_name)
         self.restore(fetch_remotes=False)  # Apply remaining patches
 
     def resolve(self):
@@ -297,12 +311,11 @@ class WorkingRepo(git.Repo):
         one time after all of the patches have been applied.
         """
         patch_name = self._resolve_conflict('resolved')
+        source_paths, parent_patch_name = self._create_patches('HEAD^')
+        if len(source_paths) > 1:
+            raise Exception("Too many patches generated")
 
-        filenames, parent_patch_name = self._create_patches('HEAD^')
-
-        source_paths = [os.path.join(self.path, f) for f in filenames]
-        self.patch_repo.add_patches(
-            [patch_name], source_paths, parent_patch_name=parent_patch_name)
+        self.patch_repo.sync_patches(source_paths, parent_patch_name)
 
         self._add_patch_annotation(patch_name)
         self.restore(fetch_remotes=False)  # Apply remaining patches
@@ -340,6 +353,21 @@ class WorkingRepo(git.Repo):
 
         return updated, removed
 
+    def _get_config(self, key):
+        try:
+            return self.config('get', config_key=key)[0]
+        except git.exc.GitException:
+            return None
+
+    def _ensure_name_and_email_set(self):
+        email = self._get_config('user.email')
+        if not email:
+            raise exc.GitConfigRequired('user.email')
+
+        name = self._get_config('user.name')
+        if not name:
+            raise exc.GitConfigRequired('user.name')
+
     def restore(self, three_way_merge=True, commit_msg=None,
                 fetch_remotes=True):
         """Applies a series of patches to the working repo's current
@@ -353,6 +381,8 @@ class WorkingRepo(git.Repo):
         # been successfully applied, skipped, or we've aborted the restore.
         #
         #####################################################################
+        self._ensure_name_and_email_set()
+
         if self.rebase_in_progress():
             raise exc.RestoreInProgress
 
@@ -391,7 +421,7 @@ class WorkingRepo(git.Repo):
                 self._update_restore_stats(delta_updated=1)
                 raise
             except git.exc.PatchAlreadyApplied:
-                self.patch_repo.remove_patches([patch_name])
+                self.patch_repo.remove_patch(patch_name)
                 self.warn("Patch '%s' appears to be upstream, removing from"
                           " patch-repo" % patch_name)
                 self._update_restore_stats(delta_removed=1)
@@ -424,7 +454,12 @@ class WorkingRepo(git.Repo):
             commit_msg = 'Refreshing patches: %d updated, %d removed' % (
                 updated, removed)
 
-        self._commit_to_patch_repo(commit_msg)
+        # Commit to patch repo
+        if not self.patch_repo.uncommitted_changes():
+            return
+        based_on = self._last_upstream_commit_hash()
+        self.patch_repo.commit(msgs=[commit_msg])
+        self.patch_repo._add_annotation('Ply-Based-On', based_on)
 
     def rollback(self, lose_uncommitted=False):
         """Rollback to that last upstream commit."""
@@ -461,19 +496,26 @@ class WorkingRepo(git.Repo):
             since, keep_subject=True, no_numbered=True, no_stat=True)
 
         # Rewrite first-line 'From <commit-hash>' -> 'From ply'
+        source_paths = []
         for filename in filenames:
-            from_filename = os.path.join(self.path, filename)
+            from_path = os.path.join(self.path, filename)
             with tempfile.NamedTemporaryFile(delete=False) as to_file:
-                with open(from_filename) as from_file:
+                with open(from_path) as from_file:
                     _fixup_patch(from_file, to_file)
 
-            shutil.move(to_file.name, from_filename)
+            # Strip 0001- prefix that git-format-patch uses
+            source_path = os.path.join(self.path, filename.split('-', 1)[1])
+            shutil.move(to_file.name, source_path)
+            os.unlink(from_path)
+
+            source_paths.append(source_path)
 
         parent_patch_name = self._get_commit_hash_and_patch_name(
             since)[1]
-        return filenames, parent_patch_name
 
-    def save(self, since=None, prefix=None):
+        return source_paths, parent_patch_name
+
+    def save(self, since=None):
         """Save a series of commits as patches into the patch-repo."""
         if self.uncommitted_changes() or self.patch_repo.uncommitted_changes():
             raise exc.UncommittedChanges
@@ -487,36 +529,10 @@ class WorkingRepo(git.Repo):
         if '..' in since:
             raise ValueError(".. not supported at the moment")
 
-        filenames, parent_patch_name = self._create_patches(since)
+        source_paths, parent_patch_name = self._create_patches(since)
 
-        patch_names = []
-        for filename in filenames:
-            # If commit already has annotation, use that patch-name
-            with open(os.path.join(self.path, filename)) as f:
-                patch_name = self._get_patch_annotation(f.read())
-
-            # Otherwise... take it from the `git format-patch` filename
-            if not patch_name:
-                # Strip 0001- prefix that git format-patch provides. Like
-                # `quilt`, `ply` uses a `series` for patch ordering.
-                patch_name = filename.split('-', 1)[1]
-
-                # Add our own subdirectory prefix, if needed
-                if prefix:
-                    patch_name = os.path.join(prefix, patch_name)
-
-            patch_names.append(patch_name)
-
-        source_paths = [os.path.join(self.path, f) for f in filenames]
-        added, updated = self.patch_repo.add_patches(
-            patch_names, source_paths, parent_patch_name=parent_patch_name)
-
-        # Remove vestigial patches (anything in the series file after the last
-        # recognized patch name)
-        series = self.patch_repo.series
-        last_idx = series.index(patch_names[-1])
-        vestigial = series[last_idx + 1:len(series)]
-        removed = self.patch_repo.remove_patches(vestigial)
+        added, updated, skipped, removed = self.patch_repo.sync_patches(
+            source_paths, parent_patch_name)
 
         # Rollback and reapply patches so that working repo has
         # patch-annotations for latest saved patches
@@ -546,7 +562,7 @@ class WorkingRepo(git.Repo):
         return self.patch_repo.check()
 
 
-class PatchRepo(git.Repo):
+class PatchRepo(Repo):
     """Represents a git repo containing versioned patch files."""
 
     def check(self):
@@ -601,10 +617,44 @@ class PatchRepo(git.Repo):
 
         self.add('series')
 
-    def add_patches(self, patch_names, source_paths, parent_patch_name=None):
-        """Add patches to the patch-repo, including add them to the series
-        file in the appropriate location.
+    def _determine_what_changed(self, source_paths, parent_patch_name):
+        added = set()
+        updated = set()
+        skipped = set()
 
+        for source_path in source_paths:
+            patch_name = os.path.basename(source_path)
+            dest_path = os.path.join(self.path, patch_name)
+            if utils.path_exists_case_sensitive(dest_path):
+                # For simplicity, we regenerate all patches, however some will
+                # be the same, so perform a file compare so we keep accurate
+                # counts of which were truly updatd
+                if utils.meaningful_diff(source_path, dest_path):
+                    updated.add(patch_name)
+                else:
+                    skipped.add(patch_name)
+            else:
+                added.add(patch_name)
+
+        # Skip all patches up to parent_patch_name
+        if parent_patch_name:
+            patch_names = self._non_recursive_series(self.series_path)
+            for patch_name in patch_names:
+                skipped.add(patch_name)
+                if patch_name == parent_patch_name:
+                    break
+
+        series = set(self.series)
+        removed = series - added - updated - skipped
+
+        return added, updated, skipped, removed
+
+    def sync_patches(self, source_paths, parent_patch_name):
+        """Sync patches into working repo, adding, updating, and removing
+        patches as necessary.
+
+        `patches` is a list of tuples representing `patch_name` and
+        `source_path`.
 
         `parent_patch_name` represents where in the `series` file we should
         insert the new patch set.
@@ -612,41 +662,43 @@ class PatchRepo(git.Repo):
         `None` indicates that the patch-set doesn't have a parent so it should
         be inserted at the beginning of the series file.
         """
-        added = set()
-        updated = set()
+        added, updated, skipped, removed = self._determine_what_changed(
+                source_paths, parent_patch_name)
 
-        # Move patches into patch-repo
-        for patch_name, source_path in zip(patch_names, source_paths):
-            # Ensure destination exists (in case a prefix was supplied)
-            dirname = os.path.dirname(patch_name)
-            dest_path = os.path.join(self.path, dirname)
-            if dirname and not os.path.exists(dest_path):
-                os.makedirs(dest_path)
+        source_lookup = {}
+        dest_lookup = {}
+        for source_path in source_paths:
+            patch_name = os.path.basename(source_path)
+            source_lookup[patch_name] = source_path
+            dest_lookup[patch_name] = os.path.join(self.path, patch_name)
 
-            dest_path = os.path.join(self.path, patch_name)
-            if os.path.exists(dest_path):
-                # For simplicity, we regenerate all patches, however some will
-                # be the same, so perform a file compare so we keep accurate
-                # counts of which were truly updatd
-                if filecmp.cmp(source_path, dest_path):
-                    os.unlink(source_path)
-                else:
-                    updated.add(patch_name)
-                    shutil.move(source_path, dest_path)
-            else:
-                added.add(patch_name)
-                shutil.move(source_path, dest_path)
+        # Toss the skipped patches
+        for patch_name in skipped:
+            if patch_name in source_lookup:
+                os.unlink(source_lookup[patch_name])
+
+        # Remove any patches that should no longer be present
+        for patch_name in removed:
+            self.rm(patch_name, force=True)
+
+        # Move the added and updated patches
+        for patch_name in added | updated:
+            shutil.move(source_lookup[patch_name],
+                        dest_lookup[patch_name])
+            self.add(patch_name)
 
         # Update series file
         with self._mutate_series_file() as entries:
+            for patch_name in removed:
+                entries.remove(patch_name)
+
             if parent_patch_name:
                 base = entries.index(parent_patch_name) + 1
             else:
                 base = 0
 
-            for idx, patch_name in enumerate(patch_names):
-                self.add(patch_name)
-
+            for idx, source_path in enumerate(source_paths):
+                patch_name = os.path.basename(source_path)
                 if patch_name in entries:
                     # Already exists, reorder patch by removing it from
                     # current location and inserting it into the new location.
@@ -654,16 +706,14 @@ class PatchRepo(git.Repo):
 
                 entries.insert(base + idx, patch_name)
 
-        return added, updated
+        return added, updated, skipped, removed
 
-    def remove_patches(self, patch_names):
-        removed = set()
+    def remove_patch(self, patch_name):
         with self._mutate_series_file() as entries:
-            for patch_name in patch_names:
-                self.rm(patch_name)
-                entries.remove(patch_name)
-                removed.add(patch_name)
-        return removed
+            # If there were any local changes and it's not in the series
+            # file, we still want to remove it, hence force=True
+            self.rm(patch_name, force=True)
+            entries.remove(patch_name)
 
     def initialize(self):
         """Initialize the patch repo (create series file and git-init)."""
@@ -674,7 +724,7 @@ class PatchRepo(git.Repo):
                 pass
 
             self.add('series')
-            self.commit('Ply init')
+            self.commit(msgs=['Ply init'])
 
     @property
     def series_path(self):
